import random

import numpy as np
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from tqdm.auto import tqdm
import re
from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn, track
import pandas as pd

def pep_gen(n:int=1000, length:int=12):
    """随机生成固定长度的多肽
    """
    peps = []
    for i in range(n):
        peps.append(''.join([random.choice(list('AGVLIFWPDEKRHSTNQYCM')) for j in range(length)]))
    return peps

def seq2record(seq_str, idx, description):
    id = str(idx) + '.' + str(description)
    return SeqRecord(Seq(seq_str), id=id)
def record_iters(seqs, idxs, descriptions):
    for seq, idx, description in zip(seqs, idxs, descriptions):
        yield seq2record(seq, idx, description)

pattern = r".*>(\d+)\.\d"
prog = re.compile(pattern)
def idx_from_line(line, prog):
    return int(prog.match(line).group(1))
def get_cluster(path, no_cluster:int = 0):
    cluster = []
    with open(path) as f:
        flag = False
        for line in f:
            if line.startswith(f">Cluster") and flag:
                break
            elif line.startswith(f">Cluster {no_cluster}"):
                    flag = True  # Startup cluster Counter
                    continue
            elif flag:
                cluster.append(idx_from_line(line, prog))
    return cluster

def cluster2df(cluster, refer_df):
    idx = pd.Index(cluster)
    return refer_df.iloc[idx].copy().reset_index(drop=True)


from abc import ABC, abstractmethod
from typing import Optional, Iterable
from rdkit import Chem


def canonize_smi(smi: str):
    try:
        return Chem.MolToSmiles(Chem.MolFromSmiles(smi), canonical=True)
    except:
        return None


def canonize_rxn(rxn: str):
    """标准化反应"""
    try:
        new_rxn = rxn.split('>>')
        r = new_rxn[0]
        p = new_rxn[1]
        new_r = '.'.join([canonize_smi(x) for x in r.split('.')])
        new_p = canonize_smi(p)
        return new_r + '>>' + new_p
    except:
        return None


def canonize_rxns(rxns: Iterable[str]):
    """批量标准化反应"""
    return [canonize_rxn(rxn) for rxn in rxns]


class SMI(object):
    def __init__(self, smi_str):
        self.smi_str = smi_str

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self.smi_str


class RXN(object):
    def __init__(self, rxn_str: str, yield_: Optional[float] = None, canonize: bool = True):
        if canonize:
            self.rxn_str = canonize_rxn(rxn_str)
        else:
            self.rxn_str = rxn_str
        if yield_:
            self.yield_ = yield_

    @property
    def products(self):
        return self.rxn_str.split('>>')[1].split('.')

    @property
    def reactants(self):
        return self.rxn_str.split('>>')[0].split('.')

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self.rxn_str



from typing import List, Iterable
from pandas import DataFrame
from .utils import split_df
from .smi import RXN


class t5chem(object):
    def __init__(self):
        pass

    def rxn2product_df(self, rxns: Iterable[RXN]) -> DataFrame:
        """将反应转化为 t5chem 的 product 预测任务"""
        products = [rxn.product for rxn in rxns]
        reactants = [rxn.reactant for rxn in rxns]
        return DataFrame({'products': products, 'reactants': reactants})



from abc import ABC, abstractmethod
import numpy as np
from typing import List


class Metrics(object):

    @staticmethod
    def remap(x: np.array, x_min, x_max):
        assert x_max > x_min, '最大最小值一样'
        return (x - x_min) / (x_max - x_min)

    @staticmethod
    def norm(scores: np.array, epsilon=0.0):
        return np.clip(Metrics.remap(scores, scores.min(), scores.max()), epsilon, 1.0)



import io
import os
import re

from pathlib import Path
from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())


def version():
    """ Get the local package version. """
    namespace = {}
    path = Path("z/__version__.py")
    exec(path.read_text(), namespace)
    return namespace["__version__"]


setup(
    name="sci-ztools",
    version=version(),
    url="https://github.com/zhaisilong/z",
    license='MIT',
    author="Zhai Silong",
    author_email="zhaisilong@outlook.com",
    description="A Unified Tool Set for Science Researches",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(where='.', exclude=('tests',), include=('*')),
    package_data={'z': ['vocab/*']},
    entry_points={
        'console_scripts': [
            'z = z.__main__:main',
        ],
    },
    install_requires=[
        'biopython',
        'rich',
        'jupyter',
        'ipykernel',
        'pandas',
        'scikit-learn',
        'numpy',
        'tqdm',
        'six',
        'future',
    ],

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.7',  # 对python的最低版本要求
)



import os
import random
from itertools import takewhile, repeat
from pathlib import Path
from typing import Union, List, Optional, Iterable
import numpy as np
import pandas as pd
import torch
from rich import console
from rich.table import Table
from sklearn.model_selection import KFold  # 交叉验证
import logging
from rich.logging import RichHandler
from logging import FileHandler
from typing import Optional
from sklearn.utils import shuffle


def get_logger(name: Optional[str] = None, filename: Optional[str] = None, level: str = 'INFO') -> logging.Logger:
    """获取一个 Rich 美化的 Logger"""
    name = name if name else __name__

    handlers = [RichHandler(
        rich_tracebacks=True,
    )]
    if filename:
        handlers.append(FileHandler(filename))

    logging.basicConfig(
        format='%(name)s: %(message)s',
        handlers=handlers)
    log = logging.getLogger(name)
    log.setLevel(level)
    return log


log = get_logger()


def read_excel(paths: Union[Path, List[Path]], drop_by: Optional[str] = None) -> pd.DataFrame:
    """读取 excel 保存为 pandas.DataFrame"""
    if isinstance(paths, List):
        # use openpyxl for better excel
        df = pd.concat([pd.read_excel(path, engine='openpyxl') for path in paths])
    elif isinstance(paths, Path):
        df = pd.read_excel(paths, engine='openpyxl')
    else:
        raise NotImplementedError
    # xlsx 的最后面往往都是空的
    # drop 保证顺序
    if drop_by:
        df.dropna(subset=[drop_by], inplace=True)
    df.reset_index(drop=True, inplace=True)
    return df


def get_device():
    "get device (CPU or GPU)"
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    n_gpu = torch.cuda.device_count()
    print("%s (%d GPUs)" % (device, n_gpu))
    return device


def iter_count(file_name):
    """获取文本行数"""
    buffer = 1024 * 1024
    with open(file_name) as f:
        buf_gen = takewhile(lambda x: x, (f.read(buffer) for _ in repeat(None)))
        return sum(buf.count('\n') for buf in buf_gen)


def df_to_table(
        pandas_dataframe: pd.DataFrame,
        rich_table: Table,
        show_index: bool = True,
        index_name: Optional[str] = None,
) -> Table:
    """Convert a pandas.DataFrame obj into a rich.Table obj.
    Args:
        pandas_dataframe (DataFrame): A Pandas DataFrame to be converted to a rich Table.
        rich_table (Table): A rich Table that should be populated by the DataFrame values.
        show_index (bool): Add a column with a row count to the table. Defaults to True.
        index_name (str, optional): The column name to give to the index column. Defaults to None, showing no value.
    Returns:
        Table: The rich Table instance passed, populated with the DataFrame values."""

    if show_index:
        index_name = str(index_name) if index_name else ""
        rich_table.add_column(index_name)

    for column in pandas_dataframe.columns:
        rich_table.add_column(str(column))

    for index, value_list in enumerate(pandas_dataframe.values.tolist()):
        row = [str(index)] if show_index else []
        row += [str(x) for x in value_list]
        rich_table.add_row(*row)

    return rich_table


def print_df(
        pandas_dataframe: pd.DataFrame,
        title: str = None,
):
    console.Console().print(df_to_table(pandas_dataframe, Table(title=title)))


def set_seeds(seed):
    "set random seeds"
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def show_ratio(df: pd.DataFrame, label='label', sort=None, n=5) -> None:
    """df 的标签中的各类比值
        Args:
            sort: 'value' or 'label'
    """
    n_all = len(df)
    if sort == 'value':
        n_classes = df[label].value_counts().reset_index().sort_values(by=label, ascending=False)
    elif sort == 'label':
        n_classes = df[label].value_counts().reset_index().sort_values(by='index')
    else:
        n_classes = df[label].value_counts().reset_index()

    n_classes = n_classes[:n]

    for i in n_classes.index:
        log.info(
            f'标签 {n_classes.at[i, "index"]} 比例为: {n_classes.at[i, label] / n_all * 100:.2f}%, 个数为: {n_classes.at[i, label]}')


def split_df(df: pd.DataFrame, shuf=True, val=True, random_state=42):
    """Split df into train/val/test set and write into files
    ratio: 8:1:1 or 9:1

    Args:
        - df (DataFrame): some data
        - shuf (bool, default=True): shuffle the DataFrame
        - val (bool, default=True): split into three set, train/val/test
    """
    if shuf:
        df = shuffle(df, random_state=random_state)

    sep = int(len(df) * 0.1)

    if val:
        test_df = df.iloc[:sep]
        val_df = df.iloc[sep:sep * 2]
        train_df = df.iloc[sep * 2:]
        return train_df, val_df, test_df
    else:
        test_df = df.iloc[:sep]
        train_df = df.iloc[sep:]
        return train_df, test_df


def kfold(df: pd.DataFrame, n_splits=5, shuffle=True, random_state=42) -> pd.DataFrame:
    """
    :param df: 输入的索引必须要对,否则会出错
    :param n_splits:
    :param shuffle:
    :param random_state:
    :return:
    """
    _df = df.copy()
    if shuffle:
        kf = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)
    else:
        kf = KFold(n_splits=n_splits)

    for fold in range(n_splits):
        _df[f'fold{fold}'] = False
    fold = 0
    for train_idxs, test_idxs in kf.split(_df):
        print(train_idxs, test_idxs)
        for i in test_idxs:
            _df.loc[i, f'fold{fold}'] = True
        fold += 1
    return _df


def get_CV(df: pd.DataFrame, n_splits=5, dir: Path = Path('CV'), header=True, index=True, cols=None):
    os.makedirs(dir, exist_ok=True)
    for fold in range(n_splits):
        _df = df.copy()
        df_fold_test = _df[_df[f'fold{fold}']]
        df_fold_train = _df[~_df[f'fold{fold}']]
        if cols:
            df_fold_test = df_fold_test[cols]
            df_fold_train = df_fold_train[cols]
            _df = _df[cols]
        fold_dir = dir / f'fold{fold}'
        os.makedirs(fold_dir, exist_ok=True)
        df_fold_test.to_csv(fold_dir / 'test.csv', header=header, index=index)
        df_fold_train.to_csv(fold_dir / 'train.csv', header=header, index=index)
        _df.to_csv(fold_dir / 'all.csv', header=header, index=index)


if __name__ == '__main__':
    df = pd.DataFrame({'a': [1, 2, 3, 4, 5, 6, 7, 1, 1], 'b': [4, 5, 6, 7, 8, 9, 10, 2, 1]})
    print(kfold(df))
